import java.util.*;
import org.apache.xerces.dom.*;
import org.w3c.dom.*;
import org.xml.sax.*;

public class DAML extends XMLIO
{
    private boolean valid; // whether a valid DAML file has been read in
    private Node root; // root of MDL document
    
    private Vector classes; // vector of all class declarations

    public DAML(String fileName, boolean test)
    {
        super(test);
        valid = false;
        root = readXMLFile(fileName); // root node of MDL file
    }

//--------------------------------------------------------------------------------------
//                              Public access methods
//--------------------------------------------------------------------------------------


    /* return the classDecl for the class with given className,
    or null if there is none. Behaves OK for a null argument,
    or if className = "Nothing". */
    public classDecl getClass(String className)
    {
        int i;
        classDecl cd,ct;
        
        cd = null;
        for (i = 0; i < classes.size(); i++)
        {
            ct = (classDecl)classes.elementAt(i);
            if (ct.className.equals(className)) {cd = ct;}
        }
        return cd;
    }
    
    // Vector of all class declarations
    public Vector classDecls() {return classes;}
            
//--------------------------------------------------------------------------------------
//                              Capture DAML from input file
//--------------------------------------------------------------------------------------

    public void captureDAML()
    {
        NodeList topNodes;
        Node topNode;
        Element el;
        String elName;
        int n,i,pass;
        String[] topChildren = {"Ontology","Class","DatatypeProperty","ObjectProperty"};
        
        
        classes = new Vector();
        valid = true;
        if (!(getLocName((Element)root).equals("RDF")))
	    {message("Root element is not an RDF element.");valid = false;}
	    else
	    {
	            
	      getRootAttributes((Element)root);
	            
	      checkChildElements((Element)root,topChildren);
	      // three-pass approach to do classes, properties, associations in that order
	      for (pass = 1; pass < 4; pass++)
	      {
	        // iterate over immediate children of the root node
	        topNodes = root.getChildNodes();
	        n = topNodes.getLength();
	        // for each child element of root element
	        if (n > 0) for (i = 0; i < n; i++)
	        {
	            topNode = topNodes.item(i);
	            if (topNode instanceof Element)
	            {
	                el = (Element) topNode;
	                elName = getLocName(el);
	                if (elName.equals("Ontology")) {doOntologyDecl(el,pass);}
	                else if (elName.equals("Class")) {doClassDecl(el,pass);}
	                else if (elName.equals("DatatypeProperty")) {doPropertyDecl(el,pass);}
	                else if (elName.equals("ObjectProperty")) {doAssociationDecl(el,pass);}
	                else {message("Cannot handle top-level element '" + elName + "'");}
	            }
	        }
	        if (pass == 1) 
	        {
	            valid = valid & checkUniqueTopClass();
	            valid = valid & checkNonCyclic();
	            if (valid) attachChildClasses();
	        }
	      }
	    }
    }
    
    /** get attributes of the top RDF node.
    These are expected to be all namespace declarations,
    and there is nothing you need to do with them.
    */
    private void getRootAttributes(Element root)
    {
        int i;
        String attName, attVal;
        NamedNodeMap attrs;
        AttrImpl att;
        attrs = root.getAttributes();
        for (i = 0; i < attrs.getLength(); i++)
        {
            att = (AttrImpl) attrs.item(i);
            attName = att.getName();
            attVal = att.getValue();
        }
    }
    
    /** handle ontology element
    nothing needs doing at present. */
    private void doOntologyDecl(Element el, int pass)
    {
        if (pass ==1)
        {
        }
    }
    
    /** handle class elements
    */
    private void doClassDecl(Element el, int pass)
    {
        classDecl cd;
        String className;
        Element label,subClassOf,comment;
        String[] classChildren = {"label","subClassOf","comment"};
        String[] classAtts = {"rdf:Class"};
        if (pass ==1)
        {
            valid = valid & checkAttributes(el,classAtts);
            valid = valid & checkChildElements(el,classChildren);
            
            label = singleChild(el,"label");
            className = getText(label);
            cd = new classDecl(className,this);
            
            subClassOf = singleChild(el,"subClassOf");
            if (subClassOf != null)
            {
                // strip off initial '#' of class resource name
                cd.superClassName = subClassOf.getAttribute("rdf:resource").substring(1);
            }
            
            comment = singleChild(el,"comment");
            if (comment != null)
            {
                cd.comment = getText(comment);
            }
            
            classes.addElement(cd);
            message("class:" + className);
        }
    }
    
    /* ensure that every class has a list of  its immediate subclasses */
    private void attachChildClasses()
    {
        int i;
        String sName;
        classDecl cd,sc;
        for (i = 0; i < classes.size(); i++)
        {
            cd = (classDecl)classes.elementAt(i);
            sName = cd.superClassName;
            if ((sName != null) &&
                !(sName.equalsIgnoreCase("nothing")))
            {
                sc = getClass(sName);
                if (sc != null) {sc.addChildClass(cd);}
            }
        }
    }
    
    /** handle DataTypeProperty ( = simple property) elements
    */
    private void doPropertyDecl(Element el, int pass)
    {
        propertyDecl pd;
        String className = "";
        String propertyName = "";
        Element label,domain,range,comment;
        String[] propertyChildren = {"label","domain","range","comment"};
        String[] propertyAtts = {"rdf:ID"};
        if (pass ==2)
        {
            valid = valid & checkAttributes(el,propertyAtts);
            valid = valid & checkChildElements(el,propertyChildren);
            label = singleChild(el,"label");
            domain = singleChild(el,"domain");
            if ((label != null) && (domain != null))
            {
                className = domain.getAttribute("rdf:resource").substring(1);
                propertyName = getText(label);
                pd = new propertyDecl(className,propertyName, this);
                range = singleChild(el,"range"); // not used yet
                comment = singleChild(el,"comment");
                if (comment != null) pd.comment = getText(comment);
                valid = valid & pd.addToClass();
            }
            else
            {
                valid = false;
                message("Property has missing property name or class name.");
            }
            message("property " + className + ":" + propertyName);
        }
    }
    
    /** handle ObjectProperty  (= association) elements
    */
    private void doAssociationDecl(Element el, int pass)
    {
        associationDecl ad;
        boolean c1u,c2u;
        String class1Name,class2Name,associationName;
        Element label,domain,range,comment;
        String[] assocChildren = {"label","domain","range","comment"};
        String[] assocAtts = {"rdf:ID"};
        if (pass ==3)
        {
            valid = valid & checkAttributes(el,assocAtts);
            valid = valid & checkChildElements(el,assocChildren);
            label = singleChild(el,"label");
            domain = singleChild(el,"domain");
            range = singleChild(el,"range");
            comment = singleChild(el,"comment");
            if ((label != null) && (domain != null) && (range != null))
            {
                class1Name = domain.getAttribute("rdf:resource").substring(1);
                class2Name = range.getAttribute("rdf:resource").substring(1);
                c1u = ("1".equals(domain.getAttribute("cardinality")));
                c2u = ("1".equals(range.getAttribute("cardinality")));
                associationName = getText(label);
                ad = new associationDecl(class1Name,associationName,class2Name,c1u,c2u,this);
                if (comment != null) ad.comment = getText(comment);
                valid = valid & ad.addToClasses();
                message("association  " + class1Name
                    + ":" + associationName + ":" + class2Name);
            }
            else
            {
                valid = false;
                message("Association has missing association name or class name.");
            }
        }
    }
    

//--------------------------------------------------------------------------------------
//                              checking methods
//--------------------------------------------------------------------------------------


/* the top of the class hierarchy is distinguished by having superClass = "Nothing"
or superClass = null. Check there is only one such class, and that all other classes
have a superclass in the hierarchy
*/
boolean checkUniqueTopClass()
{
    classDecl cd,cs;
    int i, topFound;
    String scName,topName;
    boolean res = true;
    
    topFound = 0;
    topName = "";
    for (i = 0; i < classes.size();i++)
    {
        cd = (classDecl)classes.elementAt(i);
        scName = cd.superClassName;
        if (( scName == null)|
             ((scName != null) && (scName.equalsIgnoreCase("nothing"))))
        {
            topFound++;
            if (topFound == 1) {topName = cd.className;}
            else if (topFound == 2) 
            {
                res = false;
                message("Too many top classes: " + topName
                    + ", " + cd.className);
            }
            else {message("and " + cd.className);}
        }
        else // superclass name is neither null or 'nothing' ; expect to find it
        {
            cs = getClass(scName);
            if (cs == null)
            {
                res = false;
                message("Superclass " + scName + " not found.");
            }
        }
    }
    return res;
}

/* check that the class hierarchy is non-cyclic
i.e that no class is a superclass of itself. */
boolean checkNonCyclic()
{
    classDecl cd;
    int i;
    boolean res = true;
    for (i = 0; i < classes.size(); i++)
    {
        cd = (classDecl)classes.elementAt(i);
        res = res & nonCyclic(cd,cd.className);
    }
    return res;
}

/* check that class cd has no ancestors with name cName */
boolean nonCyclic(classDecl cd, String cName)
{
    classDecl sc;
    String sName;
    boolean res;
    
    res = true;
    sName = cd.superClassName;
    if (sName == null) {res = true;}
    else if (sName.equalsIgnoreCase("nothing")) {res = true;}
    else if (sName.equals(cName)) 
    {
        message("Class " + cName + " is its own ancestor class.");
        res = false;
    }
    else
    {
        sc = getClass(sName);
        if (sc == null) 
        {
            message("Missing superclass " + sName);
            res = false;
        }
        else res = nonCyclic(sc, cName);
    }
    return res;
}

}